#include <stdlib.h>
#include <math.h>
#include "hier.h"
#include "vox.h"

/* Uncomment it for THIER */
#define	INTERPOLATE_PARENT


/***************************************************************
 * Memory management for octree.
 */
#define	HARR_NODES	100

#define	next	parent

typedef	struct	_hierchain	{
   struct _hierchain *next;
   hnode nodes[HARR_NODES];
} hierchain;

static hnode*freenode=0;
static hierchain*arr=0;
static void init_mem(void) {
   arr=0;
   freenode=0;
} /*init_mem  */
static void done_mem(void) {
   hierchain *c,*n;
   c=arr;
   while(c) {
      n=c->next;
      free((void*)c);
      c=n;
   }
   arr=0; freenode=0;
} /* done_mem */
static void add_arr(void) {
   hierchain*p=malloc(sizeof(hierchain));
   int i;
   if(!p) return;
   p->next=arr; arr=p;
   for(i=0;i<HARR_NODES;i++) {
      p->nodes[i].next=freenode;
      freenode=&p->nodes[i];
      freenode->flags=HFLAG_FREE;
   }
} /* add_arr */
static hnode* new_node(void) {
   hnode*p;
   if(!freenode) add_arr();
   p=freenode;
   if(p) { freenode=p->next; p->flags=0; p->flags=0; }
   return p;
} /* new_node */
static void del_node(hnode*p) {
   if(p) { p->next=freenode; freenode=p; }
} /* del_node */

/***************************************************************
 * Octree management.
 */
hnode *world=0;

static colorinfo c_black={0,0,0};
static livec3 zerocenter={0,0,0};
static hnode blacktransp;

static void map_dim_to_faces(int dim,int *face,int *dim1,int *dim2) {
   switch(dim) {
      case X_DIM: *face=X_SUB; *dim1=Y_SUB; *dim2=Z_SUB; break;
      case Y_DIM: *face=Y_SUB; *dim1=X_SUB; *dim2=Z_SUB; break;
      case Z_DIM: *face=Z_SUB; *dim1=X_SUB; *dim2=Y_SUB; break;
   }
} /* map_dim_to_faces */
/***************
 * Heart of the octree support.
 */
#if	0
      X
    +--+--+
    |00|01|    t(y)= (t00*t01+t10*t11)/2
 Y  +--+--+
    |10|11|    t(x)= (t00*t10+t01*t11)/2
    +--+--+
#endif
#define	mtransp(ia,ib) (long)transps[ia]*(long)transps[ib]
static transpinfo compute_transp(transpinfo transps[N_SUBS],int dir) {
   int i,j,k;
   long t;
   map_dim_to_faces(dir,&k,&i,&j);
   t=    mtransp(   0,     k) +
	 mtransp(   i,   i+k) +
	 mtransp(   j,   j+k) +
	 mtransp( i+j, i+j+k);
   return t/(TRANSPONE*4);	/* t/4/MAXTRANSPVALUE */
} /* compute_transp */
#undef	mtransp

#if	0

         | view direction
     y0 \|/
    +--+--+
->  |00|01| <-      color(x1) = ( color01*(1-t01) + t01*(color00*(1-t00) +
x0  +--+--+ x1                    color11*(1-t11) + t11*(color10*(1-t10) ) / 2
    |10|11|
    +--+--+
     y1 /|\
         |

For a cube we have to sum four colors.

#endif

typedef	struct	{
   long r,g,b;
   long tsum;
} i_ci;
#define	comp(col)	{						\
   x=((long)fwd->col*fwdti*TRANSPONE+fwdt*(bk->col*bkti));		\
   x=(long)((double)x*it);						\
   result->col+=x*tmul;						\
}
//*fwdti
static void add_colors(colorinfo *fwd,transpinfo fwdt,colorinfo *bk,transpinfo bkt,i_ci *result) {
   long fwdti=TRANSPONE-fwdt,bkti=TRANSPONE-bkt;
   long t=fwdti*TRANSPONE+fwdt*bkti;
   double it;
   long tmul=TRANSPONE*TRANSPONE-fwdt*bkt;
   long x;
   if(t<1) return;	/* Nothing to do anything */
   it=1.0/(double)t;
   comp(r); comp(g); comp(b); result->tsum+=tmul;
} /* add_colors */
#undef	comp
#define	comp(addval)	\
{ int f=fi+addval,b=bi+addval;						\
add_colors(&childs[f]->colors[direction],childs[f]->transp[dim],	\
           &childs[b]->colors[direction],childs[b]->transp[dim], &res);	\
}
static void compute_color(colorinfo *color,hnode*childs[N_SUBS],int direction) {
   i_ci res;
   int bi,fi,j,k,dim=direction>>1;
   res.r=res.g=res.b=0;
   res.tsum=0;
   map_dim_to_faces(dim,&bi,&j,&k);
   fi=0;
   if(direction&NEG_SIDE) { fi=bi; bi=0;}
   comp(0);
   comp(j);
   comp(k);
   comp(j+k);
   if(res.tsum<1) res.tsum=1;
   res.r/=(res.tsum); res.g/=(res.tsum); res.b/=(res.tsum);
   color->r=res.r;
   color->g=res.g;
   color->b=res.b;
} /* compute_color */
#undef	comp
static void update_node(hnode*p) {
   hnode *childs[N_SUBS];
   transpinfo transps[N_SUBS];
   int i;
   long t;
   do {
      for(i=0;i<N_SUBS;i++) {
         if(p->childs[i]) childs[i]=p->childs[i];
         else childs[i]=&blacktransp;	/* Better use defaults for this subspace */
      }
      /* Now every subcube information is valid */
      /* Compute transparencies */
      for(i=0;i<N_SUBS;i++) transps[i]=((childs[i])->transp[X_DIM]);
      p->transp[X_DIM]=compute_transp(transps,X_DIM);
      for(i=0;i<N_SUBS;i++) transps[i]=((childs[i])->transp[Y_DIM]);
      p->transp[Y_DIM]=compute_transp(transps,Y_DIM);
      for(i=0;i<N_SUBS;i++) transps[i]=((childs[i])->transp[Z_DIM]);
      p->transp[Z_DIM]=compute_transp(transps,Z_DIM);
      /* Compute colors */
      for(i=0;i<N_SIDES;i++) {
         compute_color(&p->colors[i],childs,i);
      }
   } while(0);
} /* update_node */
static void rec_update(hnode*p) {
   int i;
   if((p->flags&HFLAG_UPDATE)==0) return;
   for(i=0;i<N_SUBS;i++) {
      if(p->childs[i] && (p->childs[i]->flags&HFLAG_UPDATE))
         rec_update(p->childs[i]);
   }
   p->flags&=~HFLAG_UPDATE;
   update_node(p);
} /* rec_update */
void oct_update(void) {	/* update the whole tree at once */
printf("Updating...");
   rec_update(world);
printf("Updated.\n");
} /* oct_update */
static void _update_node(hnode*p) {	/* flag that we need update */
   if(p->flags&HFLAG_UPDATE) return;	/* stop if flagged */
   p->flags|=HFLAG_UPDATE;
   if(p->parent) _update_node(p->parent);
} /* _update_node */

static void init_pointers(hnode*node,hnode*parent) {
   int i;
   for(i=0;i<N_SUBS;i++) node->childs[i]=0;
   node->parent=parent;
} /* init_pointers */
static void init_node(hnode*node,colorinfo colors[N_SIDES],transpinfo transp[N_DIMS]) {
   int i;
   if(!node) return;
   for(i=0;i<N_SIDES;i++) node->colors[i]=colors[i];
   for(i=0;i<N_DIMS;i++) node->transp[i]=transp[i];
   node->flags&=~HFLAG_UPDATE;
   if(node->parent) _update_node(node->parent);
} /* init_node */
static void init_isotropic(hnode*node,colorinfo *color,transpinfo transp) {
   colorinfo colors[N_SIDES];
   transpinfo transps[N_DIMS];
   int i;
   for(i=0;i<N_SIDES;i++) colors[i]=*color;
   for(i=0;i<N_DIMS;i++) transps[i]=transp;
   init_node(node,colors,transps);
} /* init_isotropic */

int  oct_init(longint maxside) {
   longint i;
   oct_done();
   for(i=1;i<maxside;i<<=1) ;	/* power of two not less than maxside */
   init_mem();
   world=new_node();
   if(!world) return 0;
   world->size=i;
   init_pointers(world,0);
   init_isotropic(world,&c_black,TRANSPFULL);
   init_isotropic(&blacktransp,&c_black,TRANSPFULL);	/* Temporary node */
   return 1;
} /* oct_init */
void oct_done(void) {
   world=0;
   done_mem();
} /* oct_done */

static hnode* _subdivide(hnode*p,livec3*center,longint x,longint y,longint z,int step) {
   longint hs=p->size;
   int index=0;
   hnode *child;
   livec3 ncenter;
   if(x>=(*center)[X_DIM]) index|=X_SUB;
   if(y>=(*center)[Y_DIM]) index|=Y_SUB;
   if(z>=(*center)[Z_DIM]) index|=Z_SUB;
   child=p->childs[index];
   if(!child) {
      child=new_node(); if(!child) return 0;	/* alloc and check for failure */
      init_pointers(child,p);
      p->childs[index]=child; child->parent=p; child->size=hs/2;
   }
   if(hs<2) return child;	/* It's due to caller to initialize or do something with */
   ncenter[X_DIM]=(*center)[X_DIM];
   ncenter[Y_DIM]=(*center)[Y_DIM];
   ncenter[Z_DIM]=(*center)[Z_DIM];
   hs/=2;
   if(index&X_SUB) ncenter[X_DIM]+=hs; else ncenter[X_DIM]-=hs;
   if(index&Y_SUB) ncenter[Y_DIM]+=hs; else ncenter[Y_DIM]-=hs;
   if(index&Z_SUB) ncenter[Z_DIM]+=hs; else ncenter[Z_DIM]-=hs;
   return _subdivide(child,&ncenter,x,y,z,step+1);
} /* _subdivide */
static hnode* subdivide(longint x,longint y,longint z) {
//_asm int 3;
   return _subdivide(world,&zerocenter,x,y,z,0);
} /* subdivide */
int oct_addatom(longint x,longint y,longint z,colorinfo *color,transpinfo transp) {
   hnode*p=subdivide(x,y,z);
   if(!p) return 0;
   init_isotropic(p,color,transp);
   return 1;
} /* oct_addatom */

static int node_equal(hnode*n1,hnode*n2) {
   int i;
   if(n1==n2 && n1==0) return 1;
   for(i=0;i<N_DIMS;i++) {
      if(n1->transp[i]!=n2->transp[i]) return 0;
   }
   for(i=0;i<N_SIDES;i++) {
      if(n1->colors[i].r!=n2->colors[i].r) return 0;
      if(n1->colors[i].g!=n2->colors[i].g) return 0;
      if(n1->colors[i].b!=n2->colors[i].b) return 0;
   }
   return 1;
} /* node_equal */
static int _optimise(hnode*n) {
   int i,x,f=1;
   for(i=0,x=0;i<N_SUBS;i++) {
      if(n->childs[i]) {
         f&=_optimise(n->childs[i]);
         x++;
      }
   }
   if(x==i && f) {
      hnode *p=n->childs[0];
      for(i=1;i<N_SUBS;i++) {
         if(!node_equal(p,n->childs[i-1])) break;
      }
      f&=i==N_SUBS;
      if(f) {
         for(i=0;i<N_SUBS;i++) {
            del_node(n->childs[i]); n->childs[i]=0;
         }
      }
   } else
      f=x==0;	/* Node with no childs should be considered optimized */
   return f;
} /* _optimise */
int  oct_optimize(void) {
   _optimise(world);
   return 1;
} /* oct_optimise */

static void _dump(hnode*n,livec3 *coord,longint len,int step,FILE*f) {
   livec3 nc;
   int i,nn;
//if(len==4) {
   for(i=0;i<step;i++) putc(' ',f);
   printf("Node [%ld,%ld,%ld : %ld]",(*coord)[0],(*coord)[1],(*coord)[2],len);
   printf(", colors[0] : (%d,%d,%d)",n->colors[0].r,n->colors[0].g,n->colors[0].b);
   printf(", transp[] : (%d,%d,%d)",n->transp[0],n->transp[1],n->transp[2]);
   printf("\n");
//}
   for(i=0,nn=0;i<N_SUBS;i++)
      if(n->childs[i]) {
         nc[X_DIM]=(*coord)[X_DIM]; if(i&X_SUB) nc[X_DIM]+=len;
         nc[Y_DIM]=(*coord)[Y_DIM]; if(i&Y_SUB) nc[Y_DIM]+=len;
         nc[Z_DIM]=(*coord)[Z_DIM]; if(i&Z_SUB) nc[Z_DIM]+=len;
         _dump(n->childs[i],&nc,len/2,step+1,f);
         nn++;
      }
   if(nn==N_SUBS && len==1) printf("All children node!!!\n");
} /* _dump */
void oct_dump(void) {
   livec3 coord;
   longint l=world->size;
   coord[0]=coord[1]=coord[2]=-l;
   printf("\n-------------------------------\nDumping the tree:\n");
   _dump(world,&coord,l,0,stdout);
} /* oct_dump */

/***************************************************************
 * Octree visualization code.
 */
static dvec3 fwd;
static ivec3 signs;
static double ifwdsum;
static void init_view_xlate(dvec3*_fwd) {
   int i;
   double s=0;
   for(i=0;i<N_DIMS;i++) {
      double x=(*_fwd)[i],xx=fabs(x);
      fwd[i]=xx; signs[i]=x<0; s+=xx;
   }
   ifwdsum=1/s;
} /* init_view_xlate */
static void fill_colortransp(hnode*n,voxinfo*v,double rat) {
   double r,g,b,t,crat;
   hnode *p=n->parent;
   int i,j;
   if(!p) p=n;
   if(rat<0) rat=0;
   if(rat>1) rat=1.0;
   crat=1.0-rat;
   r=g=b=t=0;
   for(i=0;i<N_DIMS;i++) {
      j=i*2+signs[i];
      r+=(n->colors[j].r*rat+p->colors[j].r*crat)*fwd[i];
      g+=(n->colors[j].g*rat+p->colors[j].g*crat)*fwd[i];
      b+=(n->colors[j].b*rat+p->colors[j].b*crat)*fwd[i];
#if	0
      t+=(TRANSPFULL-n->transp[i])*fwd[i];
#elif	0
      t+=(TRANSPFULL-n->transp[i])*fwd[i]*rat+
         (TRANSPFULL-p->transp[i])*fwd[i]*crat;
#else
      t+=(n->transp[i])*fwd[i]*rat+
         (p->transp[i])*fwd[i]*crat;
#endif
   }
   r*=ifwdsum; g*=ifwdsum; b*=ifwdsum;
   t*=ifwdsum;
//if(r<0 || g<0 || b<0 || (t<0 || t>TRANSPFULL)) __emit__(0xcc);
   if(r>MAXSTOREDCOLORCOMP) r=MAXSTOREDCOLORCOMP;
   if(g>MAXSTOREDCOLORCOMP) g=MAXSTOREDCOLORCOMP;
   if(b>MAXSTOREDCOLORCOMP) b=MAXSTOREDCOLORCOMP;
   t=t; if(t<0) t=0; if(t>TRANSPFULL) t=TRANSPFULL;
   v->color.r=r; v->color.g=g; v->color.b=b; v->transp=t;
} /* fill_colortransp */
static int has_children(hnode*n) {
   int i;
   for(i=0;i<N_SUBS;i++) if(n->childs[i]) return 1;
   return 0;
} /* has_children */
static longint screenres;
static int subdivadd(hnode*n,livec3 *coord,longint len) {
   livec3 ncoord;
   int i,j,k,ix;
   int split;
   voxinfo v;
   v.size=len;
   for(i=0;i<N_DIMS;i++) {
      v.coord[i]=(*coord)[i]/*+v.size/2.0*/;
   }
   vox_check(&v,1);
   if(v.invisible && !v.split) return 1;	/* Completely invisible */
   split=v.split;	/* Compute split node criteria */
   split|=(v.z*screenres<_vox_z_one*v.size);	/* Too big on screen */
   split&=has_children(n);	/* Is there something to split? */
   if(split) {
      len/=2;
      for(k=0,ix=0;k<2;k++) {
         for(j=0;j<2;j++) {
            for(i=0;i<2;i++,ix++) {
               if(n->childs[ix]) {
                  ncoord[Z_DIM]=(*coord)[Z_DIM]+len*k;
                  ncoord[Y_DIM]=(*coord)[Y_DIM]+len*j;
                  ncoord[X_DIM]=(*coord)[X_DIM]+len*i;
                  if(!subdivadd(n->childs[ix],&ncoord,len)) return 0;
               }
            }
         }
      }
   } else if(!v.invisible) {
      double ratio;
#ifdef	INTERPOLATE_PARENT
      ratio=(_vox_z_one*v.size)/(v.z*screenres);
      if(ratio>1) ratio=1;
      ratio*=ratio*ratio;
#else
      ratio=1;
#endif
      fill_colortransp(n,&v,ratio);
      if(!vox_add(&v)) return 0;
   }
   return 1;
} /* subdivadd */
void ov_subdivide(dvec3 *_fwd,longint _screenres) {
   livec3 coord;
   int i;
   if(_screenres<1) _screenres=1;
   screenres=_screenres;
   init_view_xlate(_fwd);
   for(i=0;i<N_DIMS;i++) {
      coord[i]=-world->size;
   }
   subdivadd(world,&coord,world->size*2);
} /* ov_subdivide */
